from __future__ import annotations

import datetime
import random
import warnings
from typing import TYPE_CHECKING
from unittest.mock import Mock
from unittest.mock import call
from unittest.mock import patch
from uuid import UUID

import dash
import dash_bootstrap_components as dbc
import pytest
from dapla_metadata.datasets import ObligatoryDatasetWarning
from dapla_metadata.datasets import model
from dapla_metadata.datasets._merge import DatasetConsistencyStatus

from datadoc_editor import enums
from datadoc_editor import state
from datadoc_editor.frontend.callbacks.dataset import accept_dataset_metadata_date_input
from datadoc_editor.frontend.callbacks.dataset import accept_dataset_metadata_input
from datadoc_editor.frontend.callbacks.dataset import accept_dataset_multidropdown_input
from datadoc_editor.frontend.callbacks.dataset import open_dataset_handling
from datadoc_editor.frontend.callbacks.dataset import process_special_cases
from datadoc_editor.frontend.callbacks.dataset import remove_dataset_multidropdown_input
from datadoc_editor.frontend.callbacks.utils import MultidropdownInputTypes
from datadoc_editor.frontend.callbacks.utils import dataset_control
from datadoc_editor.frontend.constants import INVALID_DATE_ORDER
from datadoc_editor.frontend.constants import INVALID_VALUE
from datadoc_editor.frontend.fields.display_dataset import DISPLAY_DATASET
from datadoc_editor.frontend.fields.display_dataset import (
    MULTIPLE_LANGUAGE_DATASET_IDENTIFIERS,
)
from datadoc_editor.frontend.fields.display_dataset import DatasetIdentifiers

if TYPE_CHECKING:
    from dapla_metadata.datasets import Datadoc

    from datadoc_editor.frontend.callbacks.utils import MetadataInputTypes

DATASET_CALLBACKS_MODULE = "datadoc_editor.frontend.callbacks.dataset"


@pytest.fixture
def n_clicks_1():
    return 1


@pytest.fixture
def file_path():
    return "valid/path/to/file.json"


@pytest.fixture
def file_path_without_dates():
    return "valid/path/to/person_data_v1.parquet"


@pytest.mark.parametrize(
    ("metadata_identifier", "provided_value", "expected_model_value"),
    [
        (DatasetIdentifiers.SHORT_NAME, "person_data_v1", "person_data_v1"),
        (
            DatasetIdentifiers.ASSESSMENT,
            enums.Assessment.PROTECTED,
            enums.Assessment.PROTECTED.value,
        ),
        (
            DatasetIdentifiers.DATASET_STATUS,
            enums.DataSetStatus.INTERNAL,
            enums.DataSetStatus.INTERNAL.value,
        ),
        (
            DatasetIdentifiers.DATASET_STATE,
            enums.DataSetState.INPUT_DATA,
            enums.DataSetState.INPUT_DATA.value,
        ),
        (
            DatasetIdentifiers.NAME,
            "Dataset name",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Dataset name",
                    ),
                ],
            ),
        ),
        (
            DatasetIdentifiers.DESCRIPTION,
            "Dataset description",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Dataset description",
                    ),
                ],
            ),
        ),
        (
            DatasetIdentifiers.POPULATION_DESCRIPTION,
            "Population description",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Population description",
                    ),
                ],
            ),
        ),
        (DatasetIdentifiers.VERSION, 1, "1"),
        (
            DatasetIdentifiers.VERSION_DESCRIPTION,
            "Version description",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Version description",
                    ),
                ],
            ),
        ),
        (
            DatasetIdentifiers.SUBJECT_FIELD,
            "al03",
            "al03",
        ),
        (
            DatasetIdentifiers.KEYWORD,
            "one,two,three",
            ["one", "two", "three"],
        ),
        (
            DatasetIdentifiers.SPATIAL_COVERAGE_DESCRIPTION,
            "Spatial coverage description",
            model.LanguageStringType(
                [
                    model.LanguageStringTypeItem(
                        languageCode="nb",
                        languageText="Spatial coverage description",
                    ),
                    model.LanguageStringTypeItem(
                        languageCode="nn",
                        languageText="Noreg",
                    ),
                    model.LanguageStringTypeItem(
                        languageCode="en",
                        languageText="Norway",
                    ),
                ],
            ),
        ),
        (
            DatasetIdentifiers.ID,
            "2f72477a-f051-43ee-bf8b-0d8f47b5e0a7",
            UUID("2f72477a-f051-43ee-bf8b-0d8f47b5e0a7"),
        ),
        (
            DatasetIdentifiers.OWNER,
            "Seksjon for dataplattform",
            "Seksjon for dataplattform",
        ),
    ],
)
def test_accept_dataset_metadata_input_valid_data(
    metadata_identifier: DatasetIdentifiers,
    provided_value: MetadataInputTypes,
    expected_model_value: str,
    metadata: Datadoc,
):
    state.metadata = metadata
    output = accept_dataset_metadata_input(provided_value, metadata_identifier, "nb")
    assert output[0] is False
    assert output[1] == ""
    assert (
        getattr(state.metadata.dataset, metadata_identifier.value)
        == expected_model_value
    )


@pytest.mark.parametrize(
    ("metadata_identifier", "provided_value", "field", "expected_model_value"),
    [
        (
            DatasetIdentifiers.USE_RESTRICTIONS,
            "PROCESS_LIMITATIONS",
            "type",
            [
                model.UseRestrictionItem(
                    use_restriction_type="PROCESS_LIMITATIONS",
                    use_restriction_date=None,
                )
            ],
        ),
        (
            DatasetIdentifiers.USE_RESTRICTIONS,
            "2024-12-31",
            "date",
            [
                model.UseRestrictionItem(
                    use_restriction_type=None,
                    use_restriction_date=datetime.date(2024, 12, 31),
                )
            ],
        ),
    ],
)
def test_accept_dataset_multidropdown_input_valid_data(
    metadata_identifier: DatasetIdentifiers,
    provided_value: MultidropdownInputTypes,
    expected_model_value: str,
    field: str,
    metadata: Datadoc,
):
    state.metadata = metadata
    output = accept_dataset_multidropdown_input(
        provided_value, metadata_identifier, field, 0
    )
    assert output[0] is False
    assert output[1] == ""
    assert (
        getattr(state.metadata.dataset, metadata_identifier.value)
        == expected_model_value
    )


@pytest.mark.parametrize(
    ("initial_list", "index_to_remove", "expected_list"),
    [
        (
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    use_restriction_date=None,
                )
            ],
            10,
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    use_restriction_date=None,
                )
            ],
        ),
        (
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=datetime.date(2025, 9, 1),
                ),
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    use_restriction_date=datetime.date(2023, 9, 1),
                ),
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=datetime.date(2022, 9, 1),
                ),
            ],
            1,
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=datetime.date(2025, 9, 1),
                ),
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=datetime.date(2022, 9, 1),
                ),
            ],
        ),
        ([], 0, []),
        (None, 0, None),
        (
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=None,
                )
            ],
            -1,
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=None,
                )
            ],
        ),
        (
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=None,
                ),
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    use_restriction_date=None,
                ),
            ],
            0,
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    use_restriction_date=None,
                ),
            ],
        ),
        (
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=None,
                ),
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    use_restriction_date=None,
                ),
            ],
            1,
            [
                model.UseRestrictionItem(
                    use_restriction_type=enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    use_restriction_date=None,
                ),
            ],
        ),
    ],
)
def test_remove_dataset_multidropdown_input_parametrized(
    metadata: Datadoc,
    initial_list: list[model.UseRestrictionItem],
    index_to_remove: int,
    expected_list: list[model.UseRestrictionItem],
):
    state.metadata = metadata
    state.metadata.dataset.use_restrictions = initial_list
    remove_dataset_multidropdown_input(
        metadata_identifier="use_restrictions", index=index_to_remove
    )
    actual_list = state.metadata.dataset.use_restrictions
    assert actual_list == expected_list


def test_accept_dataset_metadata_input_incorrect_data_type(metadata: Datadoc):
    state.metadata = metadata
    output = accept_dataset_metadata_input(
        3.1415,
        DatasetIdentifiers.DATASET_STATE.value,
        "",
    )
    assert output[0] is True
    assert output[1] == INVALID_VALUE


earlier = str(datetime.date(2020, 1, 1))
later = str(datetime.date(2024, 1, 1))


@pytest.mark.parametrize(
    ("start_date", "end_date", "expect_error"),
    [
        (None, None, False),
        (later, None, False),
        (None, earlier, False),
        (earlier, earlier, False),
        (earlier, later, False),
        (later, earlier, True),
        (None, "12th March 1953", True),
    ],
    ids=[
        "no-input",
        "start-date-only",
        "end-date-only",
        "identical-dates",
        "correct-order",
        "incorrect-order",
        "invalid-date-format",
    ],
)
def test_accept_dataset_metadata_input_date_validation(
    metadata: Datadoc,
    start_date: str | None,
    end_date: str | None,
    expect_error: bool,  # noqa: FBT001
):
    state.metadata = metadata
    output = accept_dataset_metadata_date_input(
        DatasetIdentifiers.CONTAINS_DATA_UNTIL,
        start_date,
        end_date,
    )
    assert output[2] is expect_error
    if expect_error:
        assert output[3] == INVALID_DATE_ORDER.format(
            contains_data_from_display_name=DISPLAY_DATASET[
                DatasetIdentifiers.CONTAINS_DATA_FROM
            ].display_name,
            contains_data_until_display_name=DISPLAY_DATASET[
                DatasetIdentifiers.CONTAINS_DATA_UNTIL
            ].display_name,
        )
    else:
        assert output[1] == ""
        assert output[3] == ""


@pytest.mark.parametrize(
    "path",
    [
        "tests/resources/datasets/ifpn/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
        "  tests/resources/datasets/ifpn/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet  ",
    ],
)
@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
def test_open_dataset_handling_normal(
    open_file_mock: Mock,  # noqa: ARG001
    n_clicks_1: int,
    path: str,
):
    alert, counter = open_dataset_handling(
        n_clicks_1,
        path,
        0,
    )
    assert alert.color == "success"
    assert counter == 1


@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
def test_open_dataset_handling_file_not_found(
    open_file_mock: Mock,
    n_clicks_1: int,
    file_path: str,
):
    open_file_mock.side_effect = FileNotFoundError()

    alert, counter = open_dataset_handling(
        n_clicks_1,
        file_path,
        0,
    )
    assert alert.color == "danger"
    assert counter == dash.no_update


@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
def test_open_dataset_handling_general_exception(
    open_file_mock: Mock,
    n_clicks_1: int,
    file_path: str,
):
    open_file_mock.side_effect = ValueError()

    alert, counter = open_dataset_handling(
        n_clicks_1,
        file_path,
        0,
    )
    assert alert.color == "danger"
    assert counter == dash.no_update


@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
def test_open_dataset_handling_no_click(
    open_file_mock: Mock,  # noqa: ARG001
    file_path: str,
):
    alert, counter = open_dataset_handling(
        0,
        file_path,
        0,
    )
    assert alert
    assert counter == 1


@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
def test_open_dataset_handling_naming_standard(
    open_file_mock: Mock,  # noqa: ARG001
    n_clicks_1: int,
    file_path_without_dates: str,
):
    alert, counter = open_dataset_handling(
        n_clicks_1,
        file_path_without_dates,
        0,
    )
    assert alert.color == "warning"
    assert counter == 1


@patch(f"{DATASET_CALLBACKS_MODULE}.DaplaDatasetPathInfo")
@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
@patch(
    f"{DATASET_CALLBACKS_MODULE}.set_variables_values_inherit_dataset_derived_date_values"
)
def test_open_dataset_handling_metadata_inconsistency(
    set_vars_mock: Mock,  # noqa: ARG001
    open_file_mock: Mock,
    path_info_mock: Mock,
):
    path_info_mock.return_value.path_complies_with_naming_standard.return_value = True
    mock_metadata = Mock()
    mock_metadata.dataset_consistency_status = [
        DatasetConsistencyStatus(**status)  # type: ignore [arg-type]
        for status in [
            {"message": "Bucket name", "success": True},
            {"message": "Data product name", "success": True},
            {"message": "Dataset state", "success": False},
            {"message": "Dataset short name", "success": True},
            {"message": "Variable names", "success": True},
            {"message": "Variable datatypes", "success": True},
        ]
    ]
    open_file_mock.return_value = mock_metadata
    alert, counter = open_dataset_handling(
        n_clicks=1,
        file_path="dummy/path.parquet",
        dataset_opened_counter=0,
    )
    assert alert.color == "warning"
    assert counter == 1
    assert "Det er inkonsistens mellom data og metadata for" in str(alert)


@patch(f"{DATASET_CALLBACKS_MODULE}.DaplaDatasetPathInfo")
@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
@patch(
    f"{DATASET_CALLBACKS_MODULE}.set_variables_values_inherit_dataset_derived_date_values"
)
def test_open_dataset_handling_no_metadata_inconsistency(
    set_vars_mock: Mock,  # noqa: ARG001
    open_file_mock: Mock,
    path_info_mock: Mock,
):
    path_info_mock.return_value.path_complies_with_naming_standard.return_value = True
    mock_metadata = Mock()
    mock_metadata.dataset_consistency_status = [
        DatasetConsistencyStatus(**status)  # type: ignore [arg-type]
        for status in [
            {"message": "Bucket name", "success": True},
            {"message": "Data product name", "success": True},
            {"message": "Dataset state", "success": True},
            {"message": "Dataset short name", "success": True},
            {"message": "Variable names", "success": True},
            {"message": "Variable datatypes", "success": True},
        ]
    ]
    open_file_mock.return_value = mock_metadata
    alert, counter = open_dataset_handling(
        n_clicks=1,
        file_path="dummy/path.parquet",
        dataset_opened_counter=0,
    )
    assert alert.color == "success"
    assert counter == 1
    assert "Åpnet dataset" in str(alert)


@patch(f"{DATASET_CALLBACKS_MODULE}.DaplaDatasetPathInfo")
@patch(f"{DATASET_CALLBACKS_MODULE}.open_file")
@patch(
    f"{DATASET_CALLBACKS_MODULE}.set_variables_values_inherit_dataset_derived_date_values"
)
def test_open_dataset_alert_when_metadata_exists(
    set_vars_mock: Mock,
    open_file_mock: Mock,
    path_info_mock: Mock,
):
    set_vars_mock.assert_not_called()
    path_info_mock.return_value.path_complies_with_naming_standard.return_value = True
    parquet_path = "/tests/resources/existing_metadata_file/person_testdata_p2020-12-31_p2020-12-31_v1.parquet"
    metadata_path = "/tests/resources/existing_metadata_file/person_testdata_p2020-12-31_p2020-12-31_v1__DOC.json"
    open_file_mock.side_effect = [FileNotFoundError, Mock()]
    alert, counter = open_dataset_handling(
        n_clicks=1,
        file_path=parquet_path,
        dataset_opened_counter=0,
    )
    assert alert.color == "warning"
    assert "Datasettet finnes ikke, men metadata eksisterer" in str(alert)
    assert metadata_path in str(alert)
    assert counter is dash.no_update
    open_file_mock.assert_has_calls([call(parquet_path), call(metadata_path)])


def test_process_special_cases_keyword():
    value = "test,key,words"
    identifier = "keyword"
    expected = ["test", "key", "words"]
    assert process_special_cases(value, identifier) == expected


@patch(f"{DATASET_CALLBACKS_MODULE}.find_existing_language_string")
def test_process_special_cases_language_string(
    mock_find: Mock,
    metadata: Datadoc,
):
    state.metadata = metadata
    language = "en"
    value = "Test language string"
    identifier = random.choice(  # noqa: S311
        MULTIPLE_LANGUAGE_DATASET_IDENTIFIERS,
    )
    expected = model.LanguageStringType(
        [
            model.LanguageStringTypeItem(
                languageCode="nb",
                languageText="Existing language string",
            ),
            model.LanguageStringTypeItem(
                languageCode="en",
                languageText=value,
            ),
        ],
    )
    mock_find.return_value = expected

    assert process_special_cases(value, identifier, language) == expected


def test_process_special_cases_no_change():
    value = ["unchanged", "values"]
    identifier = "random"
    assert process_special_cases(value, identifier) == value


def test_dataset_metadata_control_return_alert(metadata: Datadoc):
    """Return alert when obligatory metadata is missing."""
    state.metadata = metadata
    missing_metadata: str
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        state.metadata.write_metadata_document()
        if issubclass(w[0].category, ObligatoryDatasetWarning):
            missing_metadata = str(w[0].message)
    result = dataset_control(missing_metadata)
    assert isinstance(result, dbc.Alert)


def test_dataset_metadata_control_not_return_alert():
    missing_metadata = None
    result = dataset_control(missing_metadata)
    assert result is None



import datetime
from dataclasses import dataclass
from unittest import mock

import dash_bootstrap_components as dbc
import pytest
from dapla_metadata.datasets import Datadoc
from dash import html
from datadoc_model.all_optional import model

from datadoc_editor import constants
from datadoc_editor import enums
from datadoc_editor import state
from datadoc_editor.enums import PseudonymizationAlgorithmsEnum
from datadoc_editor.frontend.callbacks.utils import apply_pseudonymization
from datadoc_editor.frontend.callbacks.utils import check_variable_names
from datadoc_editor.frontend.callbacks.utils import find_existing_language_string
from datadoc_editor.frontend.callbacks.utils import map_dropdown_to_pseudo
from datadoc_editor.frontend.callbacks.utils import render_multidropdown_row
from datadoc_editor.frontend.callbacks.utils import render_tabs
from datadoc_editor.frontend.callbacks.utils import save_metadata_and_generate_alerts
from datadoc_editor.frontend.callbacks.utils import update_store_data_with_inputs
from datadoc_editor.frontend.callbacks.utils import update_use_restriction_date
from datadoc_editor.frontend.callbacks.utils import update_use_restriction_type
from datadoc_editor.frontend.components.identifiers import ACCORDION_WRAPPER_ID
from datadoc_editor.frontend.components.identifiers import SECTION_WRAPPER_ID


def test_find_existing_language_string_no_existing_strings(bokmål_name: str):
    dataset_metadata = model.Dataset()
    assert find_existing_language_string(
        dataset_metadata,
        bokmål_name,
        "name",
        "nb",
    ) == model.LanguageStringType(
        [model.LanguageStringTypeItem(languageCode="nb", languageText=bokmål_name)],
    )


def test_find_existing_language_string_no_existing_strings_empty_value():
    dataset_metadata = model.Dataset()
    assert (
        find_existing_language_string(
            dataset_metadata,
            "",
            "name",
            "nb",
        )
        is None
    )


def test_find_existing_language_string_pre_existing_strings(
    english_name: str,
    bokmål_name: str,
    nynorsk_name: str,
    language_object: model.LanguageStringType,
):
    dataset_metadata = model.Dataset()
    dataset_metadata.name = language_object
    language_strings = find_existing_language_string(
        dataset_metadata,
        nynorsk_name,
        "name",
        "nn",
    )
    assert language_strings == model.LanguageStringType(
        [
            model.LanguageStringTypeItem(languageCode="en", languageText=english_name),
            model.LanguageStringTypeItem(languageCode="nb", languageText=bokmål_name),
            model.LanguageStringTypeItem(languageCode="nn", languageText=nynorsk_name),
        ],
    )


@pytest.mark.parametrize(
    ("tab", "identifier"),
    [
        ("dataset", SECTION_WRAPPER_ID),
        ("variables", ACCORDION_WRAPPER_ID),
    ],
)
def test_render_tabs(tab: str, identifier: str):
    result = render_tabs(tab)
    assert isinstance(result, html.Article)
    assert result.children[-1].id == identifier


def test_render_multidropdown_row_simple():
    row = render_multidropdown_row(
        {
            "use_restriction_type": "DELETION_ANONYMIZATION",
            "use_restriction_date": "2025-09-11",
        },
        {"component": "row_test"},
        lambda: [{"label": "Option", "value": "option"}],
        key="",
    )

    assert isinstance(row, html.Div)
    dropdown, date_input, button = row.children

    assert dropdown.value == "DELETION_ANONYMIZATION"
    assert dropdown.id == {"component": "row_test", "field": "type"}

    assert date_input.value == "2025-09-11"
    assert date_input.id == {"component": "row_test", "field": "date"}

    assert button.id == {"component": "row_test", "field": "delete"}


def test_save_and_generate_alerts():
    mock_metadata = mock.Mock()

    @dataclass
    class Variable:
        short_name: str

    mock_metadata.variables = [
        Variable(short_name="var"),
        Variable(short_name="var illegal"),
    ]
    state.metadata = mock_metadata
    result = save_metadata_and_generate_alerts(
        mock_metadata,
    )
    assert (result[1] and result[2]) is None
    assert isinstance(result[0], dbc.Alert)
    assert isinstance(result[3], dbc.Alert)


@pytest.mark.parametrize(
    ("shortname"),
    [
        ("rådyr"),
        ("Var"),
        ("Var illegal"),
        ("V"),
    ],
)
def test_illegal_shortname(shortname: str):
    @dataclass
    class MockVariable:
        short_name: str

    mock_metadata = mock.Mock(variables=[MockVariable(short_name=shortname)])
    assert isinstance(check_variable_names(mock_metadata.variables), dbc.Alert)


@pytest.mark.parametrize(
    ("shortname"),
    [
        ("var"),
        ("var1"),
        ("var_2"),
    ],
)
def test_legal_shortname(shortname: str):
    @dataclass
    class MockVariable:
        short_name: str

    mock_metadata = mock.Mock(variables=[MockVariable(short_name=shortname)])
    assert check_variable_names(mock_metadata.variables) is None


@pytest.mark.parametrize(
    ("variable", "expected_algorithm"),
    [
        (
            model.Variable(
                pseudonymization=model.Pseudonymization(
                    stable_identifier_type="FREG_SNR", encryption_algorithm="TINK-FPE"
                ),
            ),
            PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID,
        ),
        (
            model.Variable(pseudonymization=None),
            None,
        ),
        (
            model.Variable(
                pseudonymization=model.Pseudonymization(
                    encryption_algorithm="TINK-FPE"
                ),
            ),
            PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID,
        ),
        (
            model.Variable(
                pseudonymization=model.Pseudonymization(
                    encryption_algorithm="TINK-DAEAD",
                ),
            ),
            PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA,
        ),
        (
            model.Variable(
                pseudonymization=model.Pseudonymization(
                    encryption_algorithm="TINK_PPP",
                ),
            ),
            PseudonymizationAlgorithmsEnum.CUSTOM,
        ),
        (
            model.Variable(
                pseudonymization=model.Pseudonymization(
                    encryption_key_reference="custom-common-key-1",
                ),
            ),
            PseudonymizationAlgorithmsEnum.CUSTOM,
        ),
    ],
    ids=[
        "papis_with_stable_id",
        "without_pseudonymization",
        "papis_without_stable_id",
        "standard_algorithm_dapla",
        "custom_algorithm",
        "not_encryption_algorithm",
    ],
)
def test_map_dropdown_value(variable: model.Variable, expected_algorithm: str):
    assert map_dropdown_to_pseudo(variable) == expected_algorithm


@pytest.mark.parametrize(
    ("initial_value", "field", "index", "expected"),
    [
        (
            "DELETION_ANONYMIZATION",
            "type",
            0,
            model.UseRestrictionItem(
                use_restriction_type=model.UseRestrictionType.DELETION_ANONYMIZATION,
                use_restriction_date=None,
            ),
        ),
        (
            "PROCESS_LIMITATIONS",
            "type",
            0,
            model.UseRestrictionItem(
                use_restriction_type=model.UseRestrictionType.PROCESS_LIMITATIONS,
                use_restriction_date=None,
            ),
        ),
        (
            "2024-12-31",
            "date",
            1,
            model.UseRestrictionItem(
                use_restriction_type=None,
                use_restriction_date=datetime.date(2024, 12, 31),
            ),
        ),
        (
            "2025-12-31",
            "date",
            0,
            model.UseRestrictionItem(
                use_restriction_type=None,
                use_restriction_date=datetime.date(2025, 12, 31),
            ),
        ),
    ],
)
def test_update_use_restriction(initial_value, field, index, expected):
    dataset_metadata = model.Dataset()

    if index > 0:
        update_use_restriction_type(
            dataset_metadata, "DELETION_ANONYMIZATION", "use_restrictions", 0
        )

    if field == "type":
        result = update_use_restriction_type(
            dataset_metadata, initial_value, "use_restrictions", index
        )
    elif field == "date":
        result = update_use_restriction_date(
            dataset_metadata, initial_value, "use_restrictions", index
        )

    assert result[index] == expected


@pytest.mark.parametrize(
    ("type_value", "expected_type", "date_value", "expected_date"),
    [
        (
            enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
            enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
            "2025-01-01",
            "2025-01-01",
        ),
        (None, None, None, None),
    ],
)
def test_valid_type_values_update(type_value, expected_type, date_value, expected_date):
    store_data = [{"use_restriction_type": None, "use_restriction_date": None}]
    type_values = [type_value]
    date_values = [date_value]

    result = update_store_data_with_inputs(store_data, type_values, date_values)

    assert result == [
        {
            "use_restriction_type": expected_type,
            "use_restriction_date": expected_date,
        }
    ]


@pytest.mark.parametrize(
    ("store_data", "type_values", "date_values", "expected"),
    [
        (
            [
                {
                    "use_restriction_type": enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    "use_restriction_date": "2025-01-01",
                }
            ],
            [enums.UseRestrictionType.PROCESS_LIMITATIONS.value],
            ["2026-01-01"],
            [
                {
                    "use_restriction_type": enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    "use_restriction_date": "2026-01-01",
                }
            ],
        ),
        (
            [
                {
                    "use_restriction_type": enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    "use_restriction_date": "2025-01-01",
                },
                {"use_restriction_type": None, "use_restriction_date": None},
            ],
            [
                enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
            ],
            ["2025-01-01", "2026-01-01"],
            [
                {
                    "use_restriction_type": enums.UseRestrictionType.DELETION_ANONYMIZATION.value,
                    "use_restriction_date": "2025-01-01",
                },
                {
                    "use_restriction_type": enums.UseRestrictionType.PROCESS_LIMITATIONS.value,
                    "use_restriction_date": "2026-01-01",
                },
            ],
        ),
    ],
)
def test_update_store_data_with_inputs(store_data, type_values, date_values, expected):
    assert (
        update_store_data_with_inputs(store_data, type_values, date_values) == expected
    )


def test_find_existing_use_restriction_illegal_input():
    dataset_metadata = model.Dataset()
    with pytest.raises(ValueError, match="is not a valid UseRestrictionType"):
        update_use_restriction_type(
            dataset_metadata, "NOT_A_USE_RESTRICTION", "use_restrictions", 0
        )


@dataclass
class PseudoCase:
    """Test cases Pseudonymization."""

    selected_algorithm: enums.PseudonymizationAlgorithmsEnum | None
    expected_stable_type: str | None
    expected_encryption_algorithm: str | None
    expected_encryption_key_reference: str | None
    expected_algorithm_parameters: list | None
    saved_pseudonymization: model.Pseudonymization | None = None
    expected_pseudonymization_time: datetime.datetime | None = None


@pytest.mark.parametrize(
    "case",
    [
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID,
            expected_stable_type=None,
            expected_encryption_algorithm=constants.PAPIS_ALGORITHM_ENCRYPTION,
            expected_encryption_key_reference=constants.PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_algorithm_parameters=[
                {
                    constants.ENCRYPTION_PARAMETER_KEY_ID: constants.PAPIS_ENCRYPTION_KEY_REFERENCE
                },
                {
                    constants.ENCRYPTION_PARAMETER_STRATEGY: constants.ENCRYPTION_PARAMETER_STRATEGY_SKIP
                },
            ],
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID,
            expected_stable_type=constants.PAPIS_STABLE_IDENTIFIER_TYPE,
            expected_encryption_algorithm=constants.PAPIS_ALGORITHM_ENCRYPTION,
            expected_encryption_key_reference=constants.PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_algorithm_parameters=[
                {
                    constants.ENCRYPTION_PARAMETER_KEY_ID: constants.PAPIS_ENCRYPTION_KEY_REFERENCE
                },
                {
                    constants.ENCRYPTION_PARAMETER_STRATEGY: constants.ENCRYPTION_PARAMETER_STRATEGY_SKIP
                },
                {
                    constants.ENCRYPTION_PARAMETER_SNAPSHOT_DATE: datetime.datetime.now(
                        datetime.UTC
                    )
                    .date()
                    .isoformat()
                },
            ],
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA,
            expected_stable_type=None,
            expected_encryption_algorithm=constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION,
            expected_encryption_key_reference=constants.DAEAD_ENCRYPTION_KEY_REFERENCE,
            expected_algorithm_parameters=[
                {
                    constants.ENCRYPTION_PARAMETER_KEY_ID: constants.DAEAD_ENCRYPTION_KEY_REFERENCE
                },
            ],
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.CUSTOM,
            expected_stable_type=None,
            expected_encryption_algorithm=None,
            expected_encryption_key_reference=None,
            expected_algorithm_parameters=None,
        ),
        PseudoCase(
            selected_algorithm=enums.PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID,
            expected_stable_type=None,
            expected_encryption_algorithm=constants.PAPIS_ALGORITHM_ENCRYPTION,
            expected_encryption_key_reference=constants.PAPIS_ENCRYPTION_KEY_REFERENCE,
            expected_algorithm_parameters=[
                {
                    constants.ENCRYPTION_PARAMETER_KEY_ID: constants.PAPIS_ENCRYPTION_KEY_REFERENCE
                },
                {
                    constants.ENCRYPTION_PARAMETER_STRATEGY: constants.ENCRYPTION_PARAMETER_STRATEGY_SKIP
                },
            ],
            saved_pseudonymization=model.Pseudonymization(
                encryption_algorithm=constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION,
                encryption_key_reference=constants.DAEAD_ENCRYPTION_KEY_REFERENCE,
                pseudonymization_time=datetime.datetime(
                    2021, 1, 1, 0, 0, tzinfo=datetime.UTC
                ),
            ),
            expected_pseudonymization_time=datetime.datetime(
                2021, 1, 1, 0, 0, tzinfo=datetime.UTC
            ),
        ),
    ],
    ids=[
        "Selected PAPIS without stable ID",
        "Selected PAPIS with stable ID",
        "Selected DAEAD",
        "Selected custom",
        "Reselect: from DAEAD to PAPIS without stable ID",
    ],
)
def test_apply_pseudonymization_based_on_selected_algorithm(case, metadata: Datadoc):
    state.metadata = metadata
    variable = state.metadata.variables_lookup["sykepenger"]
    apply_pseudonymization(
        variable,
        case.selected_algorithm,
        case.saved_pseudonymization,
    )
    assert variable.pseudonymization is not None
    assert (
        variable.pseudonymization.encryption_algorithm
        == case.expected_encryption_algorithm
    )
    assert variable.pseudonymization.stable_identifier_type == case.expected_stable_type

    assert (
        variable.pseudonymization.encryption_key_reference
        == case.expected_encryption_key_reference
    )
    assert (
        variable.pseudonymization.encryption_algorithm_parameters
        == case.expected_algorithm_parameters
    )
    assert variable.pseudonymization.stable_identifier_type == case.expected_stable_type
    assert (
        variable.pseudonymization.pseudonymization_time
        == case.expected_pseudonymization_time
    )



SECTION_WRAPPER_ID = "section-wrapper-id"

VARIABLES_INFORMATION_ID = "variables-information"
ACCORDION_WRAPPER_ID = "accordion-wrapper"
ADD_USE_RESTRICTION_BUTTON = "add-use-restriction-button"
FORCE_RERENDER_COUNTER = "force-rerender-counter"
USE_RESTRICTION_STORE = "use-restriction-store"
USE_RESTRICTION_OPTION_STORE = "use-restriction-options-store"
USE_RESTRICTION_ID_STORE = "use-restriction-id-store"
USE_RESTRICTION_LIST_CONTAINER = "use-restriction-list-container"



import pytest

from datadoc_editor import state
from datadoc_editor.frontend.fields.display_base import DROPDOWN_DESELECT_OPTION
from datadoc_editor.frontend.fields.display_variables import get_unit_type_options
from tests.conftest import CODE_LIST_DIR
from tests.utils import TEST_RESOURCES_DIRECTORY


@pytest.mark.parametrize(
    ("code_list_csv_filepath_nb", "expected"),
    [
        (
            TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_nb.csv",
            [
                {"title": DROPDOWN_DESELECT_OPTION, "id": ""},
                {"title": "Adresse", "id": "01"},
                {"title": "Arbeidsulykke", "id": "02"},
                {"title": "Bolig", "id": "03"},
            ],
        ),
    ],
)
def test_get_unit_type_options(
    code_list_fake_structure,
    expected,
):
    state.unit_types = code_list_fake_structure
    state.unit_types.wait_for_external_result()
    assert get_unit_type_options() == expected



import pytest
import ssb_dash_components as ssb  # type: ignore[import-untyped]
from dash import html
from datadoc_model.all_optional import model

from datadoc_editor.enums import PseudonymizationAlgorithmsEnum
from datadoc_editor.frontend.callbacks.utils import map_dropdown_to_pseudo
from datadoc_editor.frontend.callbacks.utils import (
    map_selected_algorithm_to_pseudo_fields,
)
from datadoc_editor.frontend.components.builders import build_pseudo_field_section
from datadoc_editor.frontend.components.builders import (
    build_variables_pseudonymization_section,
)
from datadoc_editor.frontend.constants import PSEUDONYMIZATION

TEST_VARIABLES = [
    (
        PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID,
        model.Variable(
            short_name="helse",
            pseudonymization=model.Pseudonymization(
                stable_identifier_type="FREG_SNR", encryption_algorithm="TINK-FPE"
            ),
        ),
        2,
    ),
    (
        PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID,
        model.Variable(
            short_name="helse",
            pseudonymization=model.Pseudonymization(encryption_algorithm="TINK-FPE"),
        ),
        1,
    ),
    (
        PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA,
        model.Variable(
            short_name="helse",
            pseudonymization=model.Pseudonymization(encryption_algorithm="TINK-DAEAD"),
        ),
        1,
    ),
    (
        PseudonymizationAlgorithmsEnum.CUSTOM,
        model.Variable(
            short_name="helse",
            pseudonymization=model.Pseudonymization(encryption_algorithm="LOK"),
        ),
        5,
    ),
    (
        None,
        model.Variable(
            short_name="helse",
            pseudonymization=None,
        ),
        0,
    ),
    (
        PseudonymizationAlgorithmsEnum.CUSTOM,
        model.Variable(
            short_name="helse",
            pseudonymization=model.Pseudonymization(stable_identifier_type="FREG_SNR"),
        ),
        5,
    ),
]

TEST_IDS = [
    "papis_with_stable_id",
    "papis_without_stable_id",
    "standard_dapla",
    "custom_algorithm",
    "no_pseudonymization",
    "not_required_encryption_algorithm",
]


@pytest.mark.parametrize(
    ("expected_algorithm", "variable", "num_editable_fields"),
    TEST_VARIABLES,
    ids=TEST_IDS,
)
def test_build_variables_pseudonymization_section(
    expected_algorithm,
    variable,
    num_editable_fields,  # noqa: ARG001
):
    pseudo_section = build_variables_pseudonymization_section(
        PSEUDONYMIZATION, variable, map_dropdown_to_pseudo(variable)
    )

    assert isinstance(pseudo_section, html.Section)
    assert pseudo_section.id["title"] == PSEUDONYMIZATION

    dropdowns = [c for c in pseudo_section.children if isinstance(c, ssb.Dropdown)]
    assert len(dropdowns) == 1
    dropdown = dropdowns[0]
    assert (dropdown.value or None) == (
        expected_algorithm.value if expected_algorithm else None
    )

    container_divs = [
        c
        for c in pseudo_section.children
        if isinstance(c, html.Div) and c.id.get("variable") == variable.short_name
    ]
    assert len(container_divs) == 1


@pytest.mark.parametrize(
    ("expected_algorithm", "variable", "num_editable_fields"),
    TEST_VARIABLES,
    ids=TEST_IDS,
)
def test_build_pseudonymization_field_section(
    expected_algorithm, variable, num_editable_fields
):
    pseudo_metadata_list = map_selected_algorithm_to_pseudo_fields(expected_algorithm)

    pseudo_edit_section = build_pseudo_field_section(
        pseudo_metadata_list,
        "left",
        variable,
        variable.pseudonymization,
    )
    editable_fields = pseudo_edit_section.children

    assert len(editable_fields) == num_editable_fields, (
        f"Expected {num_editable_fields} editable fields, got {len(editable_fields)}"
    )

    assert all(isinstance(field, ssb.Input) for field in editable_fields), (
        "All editable fields should be ssb.Input components"
    )

    assert all(
        field.label == meta.display_name
        for field, meta in zip(editable_fields, pseudo_metadata_list, strict=True)
    ), "Editable field labels do not match pseudo metadata display names"



"""Functions which aren't directly called from a decorated callback."""

from __future__ import annotations

import datetime
import json
import logging
import re
import warnings
from typing import TYPE_CHECKING
from typing import cast

import arrow
import ssb_dash_components as ssb
from dapla_metadata.datasets import Datadoc
from dapla_metadata.datasets import ObligatoryDatasetWarning
from dapla_metadata.datasets import ObligatoryVariableWarning
from dapla_metadata.datasets import model
from dash import html

from datadoc_editor import config
from datadoc_editor import constants
from datadoc_editor import state
from datadoc_editor.constants import CHECK_OBLIGATORY_METADATA_DATASET_MESSAGE
from datadoc_editor.constants import CHECK_OBLIGATORY_METADATA_VARIABLES_MESSAGE
from datadoc_editor.constants import ILLEGAL_SHORTNAME_WARNING
from datadoc_editor.constants import ILLEGAL_SHORTNAME_WARNING_MESSAGE
from datadoc_editor.constants import MISSING_METADATA_WARNING
from datadoc_editor.constants import PAPIS_STABLE_IDENTIFIER_TYPE
from datadoc_editor.enums import PseudonymizationAlgorithmsEnum
from datadoc_editor.frontend.components.builders import AlertTypes
from datadoc_editor.frontend.components.builders import build_ssb_alert
from datadoc_editor.frontend.components.identifiers import ACCORDION_WRAPPER_ID
from datadoc_editor.frontend.components.identifiers import SECTION_WRAPPER_ID
from datadoc_editor.frontend.components.identifiers import VARIABLES_INFORMATION_ID
from datadoc_editor.frontend.fields.display_base import DROPDOWN_DESELECT_OPTION
from datadoc_editor.frontend.fields.display_base import MetadataMultiDropdownField
from datadoc_editor.frontend.fields.display_dataset import DISPLAY_DATASET
from datadoc_editor.frontend.fields.display_dataset import (
    OBLIGATORY_DATASET_METADATA_IDENTIFIERS_AND_DISPLAY_NAME,
)
from datadoc_editor.frontend.fields.display_dataset import DatasetIdentifiers
from datadoc_editor.frontend.fields.display_pseudo_variables import (
    OBLIGATORY_VARIABLES_METADATA_PSEUDO_IDENTIFIERS_AND_DISPLAY_NAME,
)
from datadoc_editor.frontend.fields.display_pseudo_variables import (
    PSEUDONYMIZATION_DEAD_METADATA,
)
from datadoc_editor.frontend.fields.display_pseudo_variables import (
    PSEUDONYMIZATION_METADATA,
)
from datadoc_editor.frontend.fields.display_pseudo_variables import (
    PSEUDONYMIZATION_PAPIS_WITH_STABLE_ID_METADATA,
)
from datadoc_editor.frontend.fields.display_pseudo_variables import (
    PSEUDONYMIZATION_PAPIS_WITHOUT_STABLE_ID_METADATA,
)
from datadoc_editor.frontend.fields.display_variables import (
    OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_AND_DISPLAY_NAME,
)

if TYPE_CHECKING:
    import pathlib
    from collections.abc import Callable

    import dash_bootstrap_components as dbc
    import pydantic
    from cloudpathlib import CloudPath


logger = logging.getLogger(__name__)


MetadataInputTypes = (
    str | list[str] | list[dict] | int | float | bool | datetime.date | None
)
MultidropdownInputTypes = str | None
PseudonymizationInputTypes = str | datetime.datetime | None


def _check_if_language_string_item_exists(
    language_strings: model.LanguageStringType,
    language_code: str,
) -> bool:
    if language_strings.root is None:
        return False
    return any(i.languageCode == language_code for i in language_strings.root)


def _update_language_string_item(
    language_strings: model.LanguageStringType,
    language_code: str,
    new_value: str,
) -> None:
    if language_strings.root is not None:
        for i in language_strings.root:
            if i.languageCode == language_code:
                i.languageText = new_value


def _add_language_string_item(
    language_strings: model.LanguageStringType,
    language_code: str,
    language_text: str,
) -> None:
    if language_strings.root is not None:
        language_strings.root.append(
            model.LanguageStringTypeItem(
                languageCode=language_code,
                languageText=language_text,
            ),
        )


def find_existing_language_string(
    metadata_model_object: pydantic.BaseModel,
    value: str,
    metadata_identifier: str,
    language: str,
) -> model.LanguageStringType | None:
    """Get or create a LanguageStrings object and return it."""
    language_strings = getattr(metadata_model_object, metadata_identifier)

    if language_strings is not None:
        if _check_if_language_string_item_exists(
            language_strings,
            language,
        ):
            _update_language_string_item(
                language_strings,
                language,
                value,
            )
        elif value != "":
            _add_language_string_item(
                language_strings,
                language,
                value,
            )
        else:
            return None
    elif value != "":
        language_strings = model.LanguageStringType(
            root=[
                model.LanguageStringTypeItem(
                    languageCode=language,
                    languageText=value,
                ),
            ],
        )
    else:
        # Don't create an item if the value is empty
        return None
    return language_strings


def update_use_restriction_type(
    metadata_model_object: pydantic.BaseModel,
    value: MultidropdownInputTypes,
    metadata_identifier: str,
    index: int,
) -> list[model.UseRestrictionItem]:
    """Updates the type filed on the in the multidropdown component."""
    items: list[model.UseRestrictionItem] = (
        getattr(metadata_model_object, metadata_identifier, []) or []
    )

    while len(items) <= index:
        items.append(model.UseRestrictionItem())

    items[index].use_restriction_type = (
        model.UseRestrictionType(value) if value else None
    )

    setattr(metadata_model_object, metadata_identifier, items)
    return items


def update_use_restriction_date(
    metadata_model_object: pydantic.BaseModel,
    value: MultidropdownInputTypes,
    metadata_identifier: str,
    index: int,
) -> list[model.UseRestrictionItem]:
    """Updates the date filed on the in the multidropdown component."""
    items: list[model.UseRestrictionItem] = (
        getattr(metadata_model_object, metadata_identifier, []) or []
    )

    while len(items) <= index:
        items.append(model.UseRestrictionItem())

    items[index].use_restriction_date = (
        datetime.datetime.strptime(value, "%Y-%m-%d").date() if value else None  # noqa: DTZ007
    )

    setattr(metadata_model_object, metadata_identifier, items)
    return items


def get_dataset_path() -> pathlib.Path | CloudPath | str:
    """Extract the path to the dataset from the potential sources."""
    if state.metadata.dataset_path is not None:
        return state.metadata.dataset_path
    path_from_env = config.get_datadoc_dataset_path()
    if path_from_env:
        logger.info(
            "Dataset path from env var: '%s'",
            path_from_env,
        )
    if path_from_env is None:
        path_from_env = ""
    return path_from_env


VALIDATION_ERROR = "Validation error: "
DATE_VALIDATION_MESSAGE = f"{VALIDATION_ERROR}contains_data_from must be the same or earlier date than contains_data_until"


def parse_and_validate_dates(
    start_date: str | datetime.datetime | None,
    end_date: str | datetime.datetime | None,
) -> tuple[datetime.datetime | None, datetime.datetime | None]:
    """Parse and validate the given dates.

    Validate that:
        - The dates are in YYYY-MM-DD format
        - The start date is earlier or identical to the end date.

    Examples:
    >>> parse_and_validate_dates("2021-01-01", "2021-01-01")
    (datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc))

    >>> parse_and_validate_dates("1990-01-01", "2050-01-01")
    (datetime.datetime(1990, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), datetime.datetime(2050, 1, 1, 0, 0, tzinfo=datetime.timezone.utc))

    >>> parse_and_validate_dates(None, None)
    (None, None)

    >>> parse_and_validate_dates("1st January 2021", "1st January 2021")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: Expected an ISO 8601-like string, but was given '1st January 2021'. Try passing in a format string to resolve this.

    >>> parse_and_validate_dates(datetime.datetime(2050, 1, 1, 0, 0, tzinfo=datetime.timezone.utc), "1990-01-01")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: contains_data_from must be the same or earlier date than contains_data_until

    >>> parse_and_validate_dates("2050-01-01", "1990-01-01")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: contains_data_from must be the same or earlier date than contains_data_until
    """
    parsed_start = None
    parsed_end = None
    try:
        if start_date and start_date != "None":
            parsed_start = arrow.get(start_date)
        if end_date and end_date != "None":
            parsed_end = arrow.get(end_date)
    except arrow.parser.ParserError as e:
        raise ValueError(VALIDATION_ERROR + str(e)) from e

    if parsed_start and parsed_end and (parsed_start > parsed_end):
        raise ValueError(DATE_VALIDATION_MESSAGE)

    start_output = parsed_start.astimezone(tz=datetime.UTC) if parsed_start else None
    end_output = parsed_end.astimezone(tz=datetime.UTC) if parsed_end else None

    return start_output, end_output


def render_tabs(tab: str) -> html.Article | None:
    """Render tab content."""
    if tab == "dataset":
        return html.Article(
            [
                html.Article(
                    id=SECTION_WRAPPER_ID,
                    className="workspace-content",
                ),
            ],
            className="workspace-page-wrapper",
        )
    if tab == "variables":
        return html.Article(
            [
                html.Header(
                    [
                        ssb.Paragraph(
                            id=VARIABLES_INFORMATION_ID,
                            className="workspace-info-paragraph",
                        ),
                        ssb.Input(
                            label="Filtrer",
                            searchField=True,
                            disabled=False,
                            placeholder="Variabel kortnavn...",
                            id="search-variables",
                            n_submit=0,
                            value="",
                        ),
                    ],
                    className="workspace-header",
                ),
                html.Article(
                    id=ACCORDION_WRAPPER_ID,
                    className="workspace-content",
                ),
            ],
            className="workspace-page-wrapper",
        )

    return None


def update_store_data_with_inputs(
    store_data: list[dict], type_values: list[str], date_values: list[str]
) -> list[dict]:
    """Update each use restriction in store_data with the latest type/date values."""
    for item, type_val, date_val in zip(
        store_data, type_values, date_values, strict=False
    ):
        item.update(use_restriction_type=type_val, use_restriction_date=date_val)
    return store_data


def render_multidropdown_row(
    item: dict,
    row_id: dict[str, str | int],
    options: Callable[[], list[dict[str, str]]],
    key: str | None = None,
) -> html.Div:
    """Renders a row in the multidropdown component."""
    field = cast(
        "MetadataMultiDropdownField",
        DISPLAY_DATASET[DatasetIdentifiers.USE_RESTRICTIONS],
    )

    dropdown_id = {**row_id, "field": "type"}
    date_id = {**row_id, "field": "date"}
    button_id = {**row_id, "field": "delete"}

    return html.Div(
        [
            ssb.Dropdown(
                header=field.type_display_name,
                items=options,
                placeholder=DROPDOWN_DESELECT_OPTION,
                value=item.get("use_restriction_type"),
                id=dropdown_id,
                className="dropdown-component",
                showDescription=True,
                description=field.type_description,
            ),
            ssb.Input(
                label=field.date_display_name,
                value=item.get("use_restriction_date"),
                id=date_id,
                className="input-component",
                type="date",
                showDescription=True,
                description=field.date_description,
            ),
            html.Button("", id=button_id, className="multidropdown-delete-button"),
        ],
        className="input-group-row",
        key=key,
    )


def _has_exact_word(word: str, text: str) -> bool:
    """Return True if excact word matches text."""
    return bool(re.search(rf"\b{word}\b", text))


def dataset_control(error_message: str | None) -> dbc.Alert | None:
    """Check obligatory metadata values for dataset.

    Args:
        error_message(str): A message generated by ObligatoryDatasetWarning containing names of fields missing value.
    """
    missing_metadata = [
        f[1]
        for f in OBLIGATORY_DATASET_METADATA_IDENTIFIERS_AND_DISPLAY_NAME
        if (error_message and _has_exact_word(f[0], error_message))
    ]
    if not missing_metadata:
        return None
    return build_ssb_alert(
        AlertTypes.WARNING,
        MISSING_METADATA_WARNING,
        CHECK_OBLIGATORY_METADATA_DATASET_MESSAGE,
        None,
        missing_metadata,
    )


def _parse_error_message(message: str) -> list | None:
    """Parse a string containing an error message into Python objects.

    This function processes an error message by removing predefined text and
    attempting to parse the remaining string into a list of Python objects.

    Args:
        message (str): The error message string to be parsed.

    Returns:
        A list of parsed objects if the string can be successfully parsed.
        returns None if the string is empty after processing and a empty list
        if the string cannot be parsed into JSON.
    """
    parsed_string = message.replace("Obligatory metadata is missing: ", "").strip()
    parsed_string = parsed_string.replace("'", '"')
    if not parsed_string:
        return None
    try:
        # Attempt to parse the JSON string
        return json.loads(parsed_string)
    except json.JSONDecodeError:
        logger.exception("Error parsing JSON {e}")
        return []


def _get_dict_by_key(
    metadata_list: list[dict[str, list[str]]],
    key: str,
) -> dict[str, list[str]] | None:
    """Return the first dictionary containing the specified key.

    This function searches through a list of dictionaries and returns the first
    dictionary that contains the specified key.

    Args:
        metadata_list: A list of dictionaries to search.
        key: The key to search for in each dictionary.

    Returns:
        The first dictionary containing the specified key,
        or None if no such dictionary is found.

    """
    return next((item for item in metadata_list if key in item), None)


def variables_control(
    error_message: list[str] | None, variables: list
) -> dbc.Alert | None:
    """Check obligatory metadata for variables and return an alert if any metadata is missing.

    This function parses an error message to identify missing obligatory metadata
    fields for variables. If missing metadata is found, it generates an alert.

    Args:
        error_message: A message generated by ObligatoryVariableWarning
            containing the variable short name and a list of field names with missing values.
        variables: list of datadoc variables

    Returns:
        An alert object if there are missing metadata fields, otherwise None.
    """
    missing_metadata: list = []
    # NOTE(tilen1976): This is a bug fix - can be fragile. The bug is probably cause by multiple validations which are triggered by "write_metadata_document()"
    # Use first element in variables error messages because there is a second validation which overwrites the result
    error_message_parsed = (
        _parse_error_message(str(error_message[0])) if error_message else None
    )
    obligatory_fields = (
        OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_AND_DISPLAY_NAME
        + OBLIGATORY_VARIABLES_METADATA_PSEUDO_IDENTIFIERS_AND_DISPLAY_NAME
    )
    for variable in variables:
        if error_message_parsed:
            fields_by_variable = _get_dict_by_key(
                error_message_parsed,
                variable.short_name,
            )
            if fields_by_variable is not None:
                missing_metadata_field = [
                    f[1]
                    for f in obligatory_fields
                    if error_message and f[0] in fields_by_variable[variable.short_name]
                ]
                missing_metadata_fields_to_string = ", ".join(missing_metadata_field)
                missing_metadata.append(
                    f"{variable.short_name}: {missing_metadata_fields_to_string}",
                )
    if not missing_metadata:
        return None
    return build_ssb_alert(
        AlertTypes.WARNING,
        MISSING_METADATA_WARNING,
        CHECK_OBLIGATORY_METADATA_VARIABLES_MESSAGE,
        None,
        missing_metadata,
    )


def check_variable_names(
    variables: list,
) -> dbc.Alert | None:
    """Checks if a variable shortname complies with the naming standard.

    Returns:
        An ssb alert with a message saying that what names dont comply with the naming standard.
    """
    illegal_names: list = [
        v.short_name
        for v in variables
        if not re.match(r"^[a-z0-9_]{2,}$", v.short_name)
    ]

    if not illegal_names:
        return None
    return build_ssb_alert(
        AlertTypes.WARNING,
        ILLEGAL_SHORTNAME_WARNING,
        ILLEGAL_SHORTNAME_WARNING_MESSAGE,
        None,
        illegal_names,
    )


def save_metadata_and_generate_alerts(metadata: Datadoc) -> list:
    """Save the metadata document to disk and check obligatory metadata.

    Returns:
        List of alerts including obligatory metadata warnings if missing,
        and success alert if metadata is saved correctly.
    """
    missing_obligatory_dataset = ""
    # Use list so the first error message is not overwritten
    missing_obligatory_variables = []

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        try:
            metadata.write_metadata_document()
            success_alert = build_ssb_alert(
                AlertTypes.SUCCESS,
                "Lagret metadata",
            )
            for warning in w:
                if issubclass(warning.category, ObligatoryDatasetWarning):
                    missing_obligatory_dataset = str(warning.message)
                elif issubclass(warning.category, ObligatoryVariableWarning):
                    missing_obligatory_variables.append(str(warning.message))
                else:
                    logger.warning(
                        "An unexpected warning was caught: %s",
                        warning.message,
                    )
        except (ValueError, FileNotFoundError) as e:
            if isinstance(e, ValueError):
                logger.exception("Unable to save metadata document, no dataset found")
            elif isinstance(e, FileNotFoundError):
                logger.exception(
                    "Unable to save metadata document, file %s not found",
                    metadata.dataset_path,
                )
            success_alert = build_ssb_alert(
                AlertTypes.ERROR,
                "Kunne ikke lagre metadata",
            )
    logger.debug(
        "Missing variables obligatory fields: %s", missing_obligatory_variables
    )
    return [
        success_alert,
        dataset_control(missing_obligatory_dataset),
        variables_control(missing_obligatory_variables, metadata.variables),
        check_variable_names(metadata.variables),
    ]


def map_selected_algorithm_to_pseudo_fields(
    selected_algorithm: PseudonymizationAlgorithmsEnum | None,
) -> list:
    """Map a PseudonymizationAlgorithms enum value to the correct pseudonymization input list.

    Examples:
    >>> pseudo_fields = map_selected_algorithm_to_pseudo_fields(PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID)
    >>> len(pseudo_fields)
    1
    >>> pseudo_fields = map_selected_algorithm_to_pseudo_fields(PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID)
    >>> len(pseudo_fields)
    2
    >>> pseudo_fields = map_selected_algorithm_to_pseudo_fields(PseudonymizationAlgorithmsEnum.CUSTOM)
    >>> len(pseudo_fields)
    5
    """
    mapping = {
        PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID: PSEUDONYMIZATION_PAPIS_WITHOUT_STABLE_ID_METADATA,
        PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID: PSEUDONYMIZATION_PAPIS_WITH_STABLE_ID_METADATA,
        PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA: PSEUDONYMIZATION_DEAD_METADATA,
        PseudonymizationAlgorithmsEnum.CUSTOM: PSEUDONYMIZATION_METADATA,
    }

    if selected_algorithm is None:
        return []

    return mapping.get(selected_algorithm, [])


def map_dropdown_to_pseudo(
    variable: model.Variable,
) -> PseudonymizationAlgorithmsEnum | None:
    """Return dropdown algorithm value for a variable's pseudonymization."""
    if variable.pseudonymization:
        match variable.pseudonymization.encryption_algorithm:
            case constants.PAPIS_ALGORITHM_ENCRYPTION:
                if (
                    variable.pseudonymization.stable_identifier_type
                    == PAPIS_STABLE_IDENTIFIER_TYPE
                ):
                    return PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID
                return PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID
            case constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION:
                return PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA
            case _:
                return PseudonymizationAlgorithmsEnum.CUSTOM
    return None


def apply_pseudonymization(
    variable: model.Variable,
    selected_algorithm: PseudonymizationAlgorithmsEnum,
    transfer_pseudonymzation: model.Pseudonymization | None,
) -> None:
    """Apply a pseudonymization algorithm to a variable and update its metadata.

    Depending on the selected algorithm, this function creates and assigns a
    corresponding `Pseudonymization` object. If previously stored values exist
    for algorithm-specific optional fields, they are transferred to the new
    `Pseudonymization`. If the selected algorithm is `CUSTOM`, all available
    values are transferred.

    Args:
        variable (model.Variable): The variable to pseudonymize.
        selected_algorithm (PseudonymizationAlgorithmsEnum): The pseudonymization algorithm to apply.
        transfer_pseudonymzation (model.Pseudonymization | None): Existing
            pseudonymization data to transfer, if applicable.

    Returns:
        None
    """
    if variable.short_name:
        match selected_algorithm:
            case PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITHOUT_STABLE_ID:
                state.metadata.add_pseudonymization(
                    variable.short_name,
                    model.Pseudonymization(
                        encryption_algorithm=constants.PAPIS_ALGORITHM_ENCRYPTION,
                        pseudonymization_time=transfer_pseudonymzation.pseudonymization_time
                        if transfer_pseudonymzation
                        else None,
                    ),
                )
            case PseudonymizationAlgorithmsEnum.PAPIS_ALGORITHM_WITH_STABLE_ID:
                state.metadata.add_pseudonymization(
                    variable.short_name,
                    model.Pseudonymization(
                        encryption_algorithm=constants.PAPIS_ALGORITHM_ENCRYPTION,
                        stable_identifier_type=constants.PAPIS_STABLE_IDENTIFIER_TYPE,
                        pseudonymization_time=transfer_pseudonymzation.pseudonymization_time
                        if transfer_pseudonymzation
                        else None,
                        stable_identifier_version=transfer_pseudonymzation.stable_identifier_version
                        if transfer_pseudonymzation
                        else None,
                    ),
                )
            case PseudonymizationAlgorithmsEnum.STANDARD_ALGORITM_DAPLA:
                state.metadata.add_pseudonymization(
                    variable.short_name,
                    model.Pseudonymization(
                        encryption_algorithm=constants.STANDARD_ALGORITM_DAPLA_ENCRYPTION,
                        pseudonymization_time=transfer_pseudonymzation.pseudonymization_time
                        if transfer_pseudonymzation
                        else None,
                    ),
                )
            case PseudonymizationAlgorithmsEnum.CUSTOM:
                state.metadata.add_pseudonymization(
                    variable.short_name,
                    pseudonymization=model.Pseudonymization(
                        encryption_algorithm=None,
                        encryption_key_reference=None,
                        pseudonymization_time=transfer_pseudonymzation.pseudonymization_time
                        if transfer_pseudonymzation
                        else None,
                        stable_identifier_type=None,
                        stable_identifier_version=None,
                        encryption_algorithm_parameters=None,
                    ),
                )


def parse_and_validate_pseudonymization_time(
    pseudo_date: str | datetime.datetime | None,
) -> datetime.datetime | None:
    """Parse and validate the given date.

    Validate that:
        - The date is in YYYY-MM-DD format

    Examples:
    >>> parse_and_validate_pseudonymization_time("2021-01-01")
    datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)

    >>> parse_and_validate_pseudonymization_time("1990-01-01")
    datetime.datetime(1990, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)

    >>> parse_and_validate_pseudonymization_time("1st January 2021")
    Traceback (most recent call last):
    ...
    ValueError: Validation error: Expected an ISO 8601-like string, but was given '1st January 2021'. Try passing in a format string to resolve this.

    >>> parse_and_validate_pseudonymization_time(None)
    None
    """
    if pseudo_date is None:
        return None
    parsed_date = None
    try:
        if pseudo_date:
            parsed_date = arrow.get(pseudo_date)
    except arrow.parser.ParserError as e:
        raise ValueError(VALIDATION_ERROR + str(e)) from e

    return parsed_date.astimezone(tz=datetime.UTC) if parsed_date else None


def update_selected_pseudonymization(
    variable: model.Variable,
    old_algorithm: PseudonymizationAlgorithmsEnum,
    new_algorithm: PseudonymizationAlgorithmsEnum,
) -> None:
    """Update the pseudonymization algorithm for a variable.

    This function replaces the variable's existing pseudonymization with a new one:
        - Removes the pseudonymization defined by the old algorithm.
        - Transfers old pseudonymization values to 'apply_pseudonymization'
        - Applies a new pseudonymization based on the newly selected algorithm.

    Args:
        variable (model.Variable):
            The variable whose pseudonymization is being updated.
        old_algorithm (PseudonymizationAlgorithmsEnum):
            The previously applied pseudonymization algorithm.
        new_algorithm (PseudonymizationAlgorithmsEnum):
            The newly selected pseudonymization algorithm.
    """
    if variable.short_name:
        transfer_pseudonymization = (
            variable.pseudonymization if variable.pseudonymization else None
        )
        state.metadata.remove_pseudonymization(variable.short_name)
        logger.debug(
            "Updating pseuonymization step 1: Remove pseudonymization for %s",
            variable.short_name,
        )
        apply_pseudonymization(variable, new_algorithm, transfer_pseudonymization)
        logger.debug(
            "Updating pseuonymization step 2: Add new pseudonymization for %s.",
            variable.short_name,
        )
        logger.info(
            "Updating pseudonymization algorithm for %s from %s to %s.",
            variable.short_name,
            old_algorithm,
            new_algorithm,
        )


def delete_pseudonymization(variable: model.Variable) -> None:
    """Remove pseudonymization for a Variable.."""
    state.metadata.remove_pseudonymization(
        variable.short_name,
    ) if variable.short_name else logger.debug(
        "Could not delete pseudonymization for %s", variable.short_name
    )
    logger.info("Removed pseudonymization for %s", variable.short_name)
